(ns com.aurellem.world.practice
 (:use (com.aurellem.gb saves util constants gb-driver vbm items assembly characters))
 (:use (com.aurellem.exp pokemon))
 (:use (com.aurellem.exp item-bridge))
 (:import [com.aurellem.gb.gb_driver SaveState]))


;;(def original-rom (rom(root)))


(def hex-pc (comp hex PC))

(defn nstep [state n]
  (if (zero? n) state
      (recur (step state) (dec n))))


(defn view-memory*
  "View a region of indexable memory in the given state."
  [state start length]
  ((comp vec map)
   #((comp  aget) (memory state) %)
   (range start (+ start length))))


(defn state-surprise
  "This is one tick before the trainer goes [!]"
  []
  (->
   (pre-trainer-battle)
   (step [:r])
   (step)
   (ntick 88147)
   ;(step [:r])

   ;(step [:r])
   ;(step [:r])
   ;(step [:r])
   ;(step [:r])
   ))

(defn state-inject
  "I have replaced the letter e with e-acute @ 0xC4E8."
  []
  (read-state "inject-surprise"))




(defn pc-trail
  "Track the PC for a number of ticks."
  [state ticks]
  (tick state)
  (set-state! state)
  (loop [pcs [(PC)] ]
    (if (> (count pcs) ticks) pcs
        (do
          (com.aurellem.gb.Gb/tick)
          (recur (conj pcs (PC)))))))



(defn differences [list-1 list-2]
  (remove
   (fn [[a b c]] (= b c))
   (map vector
        (range)
        list-1
        list-2)))

(defn pc-diff [state-1 state-2]
  (differences (map hex (pc-trail state-1 10000))
               (map hex (pc-trail state-2 10000))))


(defn memory-diff [state-1 state-2]
  (remove
  (fn[[a b c]] (= b c))
  (map (comp vec (partial map hex) list)
       (range)
       (vec (memory state-1))
       (vec (memory state-2)))
   ))



(defn state-speak
  "This is when the trainer speaks."
  []
  (->
   (pre-trainer-battle)
   (set-memory 0xD354 0x0)
   (step [:r])
   (step)
   (ntick 88147)
   (tick)
   (nstep 102)

   ;(step [:r])

   ;(step [:r])
   ;(step [:r])
   ;(step [:r])
   ;(step [:r])
   ))
  
  

(defn get-memory [state n]
  (aget (memory state) n))

(defn first-change
  "Watch the current memory location as it ticks,
return the first state that differs at location mem." 
  [state n]
  (tick state)
  (set-state! state)
  (let [init (aget (memory state) n)]
    (loop []
      (if (= (aget (memory) n) init)
        (do
          (com.aurellem.gb.Gb/tick)
          (recur))))
    (update-state)))




(defn spell-array
  [array mem n]
  (character-codes->str
          (take n (drop mem
                        (vec array)))))

(defn spell
  ([state mem n]
     (spell (memory state) mem n))
  ([mem n] (spell @current-state mem n)))



(comment
  (println)
  (print (character-codes->str (take 6000 (drop 0xA75F4
                                               (vec(com.aurellem.gb.gb-driver/rom)))))))

;(dorun (map println (view-memory* (state-surprise)  0x1AEF 1600)))



(def surprise-words
  [0x80 0xAB 0xAB 0x7F 0xB1 0xA8 0xA6 0xA7 0xB3 0xE7 0x7F 0x8B 0xA4 0xB3 0xE0 0xB2 0x7F 0xB1 0xAE 0xAB 0xAB 0x7F 0xB3 0xA7 0xA4 0x7F 0xA3 0xA8 0xA2 0xA4 0xE7])

(defn sublist
  "Unshifts the list until the sublist is at the start."
  [list sub]
  (cond
    (empty? sub) list
    (empty? list) nil
    (= (take (count sub) list) sub) list
    :else (recur (rest list) sub)))

(defn find-sublist
  "Returns the position of the first occurence of sublist."
  [list sub]
  (loop [n 0 a list]
    (cond
      (empty? a) nil
      (= (take (count sub) a) sub) n
      :else (recur (inc n) (rest a)))))

(defn find-sublists
  "Returns a vector of the occurences of sublists."
  [list sub]
  (let [m (find-sublist list sub)]
    (if (nil? m) '()
        (cons m
              (map (partial + (inc m))
                   (find-sublists
                    (drop (inc m) list)
                    sub))))))



(defn search-rom
  "Search for the given codes in ROM, returning short snippets of
text around the results."
  ([codes k]
     (search-rom com.aurellem.gb.gb-driver/original-rom codes k))
  ([rom codes k]
     (map
      (fn [n]
        [(hex n)
         (take k (drop n rom))])
      
      (find-sublists
       rom
       codes))))

(defn spelling-bee
  "Search for the given string in ROM, returning short snippets of
  text around the results."
  ([str k]
     (spelling-bee com.aurellem.gb.gb-driver/original-rom str k))
  ([rom str k]
     (map
      (fn [[address snip]]
        [address (character-codes->str snip)])
        (search-rom rom (str->character-codes str) k))))
     




(defn change-speech
  ([state str k]
     (loop [ops str
            s state
            n k]
       (if (empty? ops) s
           (recur
            (rest ops)
            (set-memory (first-change s (+ 0xC4B9 n)) (+ 0xC4B9 n)
                        (first ops))
            (if (not= n 19) (inc n)
                (+ n 21))))))
  ([str k]
     (change-speech (state-speak) str k))
  ([str]
     (change-speech str 0)))



(defn rewrite-memory
  "Alters the vector of memory. Treats strings as lists of character
ops."
  ([mem start strs-or-ops]
     (let [x (first strs-or-ops)]
       (cond (empty? strs-or-ops) mem
             (string? x)
             
             (recur mem start
                    (concat
                     (str->character-codes x)
                     (rest strs-or-ops)))
        :else
        (recur
         (assoc mem start x)
         (inc start)
         (rest strs-or-ops))))))

(def rewrite-rom
  "Alters the ROM array using write-memory. Takes a list of
various strings/bytes as data."
  (partial rewrite-memory (vec (rom(root)))))

(defn restore-rom! [] (write-rom! original-rom))



(def oak-intro
  
                  (list "Hello there!"
                        0x4F
                        "Welcome to the"
                        0x55
                        "world of "
                        0x54
                        "MON!"
                        0x51
                        "My name is OAK!"
                        0x4F
                        "People call me"
                        0x55
                        "the "
                        0x54
                        "MON PROF!"
                        0x58
                        "This world is"
                        0x4F
                        "inhabited by"
                        0x55
                        "creatures called"
                        0x55
                        0x54
                        "MON!"
                        0x50
                        0x50
                        0x51
                        "For some people, "
                        0x4F
                        0x54
                        "MON are"
                        0x55
                        "pets. Others use"
                        0x55
                        "them for fights."
                        0x51
                        "Myself..."
                        0x51
                        "I study "
                        0x54
                        "MON"
                        0x4F
                        "as a profession."
                        0x58
                        "First,  what is"
                        0x4F
                        "your name?"
                        0x58
                        "This is my grand-"
                        0x4F
                        "son. He's been"
                        0x55
                        "your rival since"
                        0x55
                        "you were a baby."
                        0x51
                        "...Erm,  what is"
                        0x4F
                        "his name again?"
                        0x58
                        0x52
                        "!"
                        0x51
                        "Your very own"
                        0x4F
                        0x54
                        "MON legend is"
                        0x55
                        "about to unfold!"
                        0x51
                        "A world of dreams"
                        0x4F
                        "and adventures"
                        0x55
                        "with "
                        0x54
                        "MON"
                        0x55
                        "awaits! Let's go!"
                        0x57)                 )

(defn state-intro
  "Professor Oak speaks!"
  []
  (do
    (->
     (rewrite-rom 0xA07BB
                  (list 0x87 0xA4 0xAB 0xAB 0xAE 0x7F 0xB3 0xA7 0xA4 0xB1 0xA4 0xE7 0x4F 0x96 0xA4 0xAB 0xA2 0xAE 0xAC 0xA4 0x7F 0xB3 0xAE 0x7F 0xB3 0xA7 0xA4 0x55 0xB6 0xAE 0xB1 0xAB 0xA3 0x7F 0xAE 0xA5 0x7F 0x54 0x8C 0x8E 0x8D 0xE7 0x51 0x8C 0xB8 0x7F 0xAD 0xA0 0xAC 0xA4 0x7F 0xA8 0xB2 0x7F 0x8E 0x80 0x8A 0xE7 0x4F 0x8F 0xA4 0xAE 0xAF 0xAB 0xA4 0x7F 0xA2 0xA0 0xAB 0xAB 0x7F 0xAC 0xA4 0x55 0xB3 0xA7 0xA4 0x7F 0x8F 0x8C 0x8E 0x8D 0x7F 0x8F 0x91 0x8E 0x85 0xE7 0x58 0x93 0xA7 0xA8 0xB2 0x7F 0xB6 0xAE 0xB1 0xAB 0xA3 0x7F 0xA8 0xB2 0x4F 0xA8 0xAD 0xA7 0xA0 0xA1 0xA8 0xB3 0xA4 0xA3 0x7F 0xA1 0xB8 0x55 0xA2 0xB1 0xA4 0xA0 0xB3 0xB4 0xB1 0xA4 0xB2 0x7F 0xA2 0xA0 0xAB 0xAB 0xA4 0xA3 0x55 0x54 0x8C 0x8E 0x8D 0xE7 0x50 0x50 0x51 0x85 0xAE 0xB1 0x7F 0xB2 0xAE 0xAC 0xA4 0x7F 0xAF 0xA4 0xAE 0xAF 0xAB 0xA4 0xF3 0x7F 0x4F 0x54 0x8C 0x8E 0x8D 0x7F 0xA0 0xB1 0xA4 0x55 0xAF 0xA4 0xB3 0xB2 0xF1 0x7F 0x8E 0xB3 0xA7 0xA4 0xB1 0xB2 0x7F 0xB4 0xB2 0xA4 0x55 0xB3 0xA7 0xA4 0xAC 0x7F 0xA5 0xAE 0xB1 0x7F 0xA5 0xA8 0xA6 0xA7 0xB3 0xB2 0xF1 0x51 0x8C 0xB8 0xB2 0xA4 0xAB 0xA5 0xF1 0xF1 0xF1 0x51 0x88 0x7F 0xB2 0xB3 0xB4 0xA3 0xB8 0x7F 0x54 0x8C 0x8E 0x8D 0x4F 0xA0 0xB2 0x7F 0xA0 0x7F 0xAF 0xB1 0xAE 0xA5 0xA4 0xB2 0xB2 0xA8 0xAE 0xAD 0xF1 0x58 0x85 0xA8 0xB1 0xB2 0xB3 0xF3 0x7F 0x7F 0xB6 0xA7 0xA0 0xB3 0x7F 0xA8 0xB2 0x4F 0xB8 0xAE 0xB4 0xB1 0x7F 0xAD 0xA0 0xAC 0xA4 0xE6 0x58 0x93 0xA7 0xA8 0xB2 0x7F 0xA8 0xB2 0x7F 0xAC 0xB8 0x7F 0xA6 0xB1 0xA0 0xAD 0xA3 0xE3 0x4F 0xB2 0xAE 0xAD 0xF1 0x7F 0x87 0xA4 0xE0 0xB2 0x7F 0xA1 0xA4 0xA4 0xAD 0x55 0xB8 0xAE 0xB4 0xB1 0x7F 0xB1 0xA8 0xB5 0xA0 0xAB 0x7F 0xB2 0xA8 0xAD 0xA2 0xA4 0x55 0xB8 0xAE 0xB4 0x7F 0xB6 0xA4 0xB1 0xA4 0x7F 0xA0 0x7F 0xA1 0xA0 0xA1 0xB8 0xF1 0x51 0xF1 0xF1 0xF1 0x84 0xB1 0xAC 0xF3 0x7F 0x7F 0xB6 0xA7 0xA0 0xB3 0x7F 0xA8 0xB2 0x4F 0xA7 0xA8 0xB2 0x7F 0xAD 0xA0 0xAC 0xA4 0x7F 0xA0 0xA6 0xA0 0xA8 0xAD 0xE6 0x58 0x52 0xE7 0x51 0x98 0xAE 0xB4 0xB1 0x7F 0xB5 0xA4 0xB1 0xB8 0x7F 0xAE 0xB6 0xAD 0x4F 0x54 0x8C 0x8E 0x8D 0x7F 0xAB 0xA4 0xA6 0xA4 0xAD 0xA3 0x7F 0xA8 0xB2 0x55 0xA0 0xA1 0xAE 0xB4 0xB3 0x7F 0xB3 0xAE 0x7F 0xB4 0xAD 0xA5 0xAE 0xAB 0xA3 0xE7 0x51 0x80 0x7F 0xB6 0xAE 0xB1 0xAB 0xA3 0x7F 0xAE 0xA5 0x7F 0xA3 0xB1 0xA4 0xA0 0xAC 0xB2 0x4F 0xA0 0xAD 0xA3 0x7F 0xA0 0xA3 0xB5 0xA4 0xAD 0xB3 0xB4 0xB1 0xA4 0xB2 0x55 0xB6 0xA8 0xB3 0xA7 0x7F 0x54 0x8C 0x8E 0x8D 0x55 0xA0 0xB6 0xA0 0xA8 0xB3 0xB2 0xE7 0x7F 0x8B 0xA4 0xB3 0xE0 0xB2 0x7F 0xA6 0xAE 0xE7 0x57 0x0 0x83 0xAE 0x7F 0xB8 0xAE 0xB4 0x7F 0xB6 0xA0 0xAD 0xB3 0x7F 0xB3 0xAE 0x4F 0xA6 0xA8 0xB5 0xA4 0x7F 0xA0 0x7F 0xAD 0xA8 0xA2 0xAA 0xAD 0xA0 0xAC 0xA4 0x55 0xB3 0xAE 0x7F 0x50 0x1 0x6D 0xCD 0x0 0xE6 0x57 0x0 0x91 0xA8 0xA6 0xA7 0xB3 0xE7 0x7F 0x92 0xAE 0x7F 0xB8 0xAE 0xB4 0xB1 0x4F 0xAD 0xA0 0xAC 0xA4 0x7F 0xA8 0xB2 0x7F 0x52 0xE7 0x58 0x0 0x93 0xA7 0xA0 0xB3 0xBD 0x7F 0xB1 0xA8 0xA6 0xA7 0xB3 0xE7 0x7F 0x88 0x4F 0xB1 0xA4 0xAC 0xA4 0xAC 0xA1 0xA4 0xB1 0x7F 0xAD 0xAE 0xB6 0xE7 0x7F 0x87 0xA8 0xB2 0x55 0xAD 0xA0 0xAC 0xA4 0x7F 0xA8 0xB2 0x7F 0x53 0xE7 0x58 0x1 0x3F 0xCD 0x0 0x7F 0xA0 0xAD 0xA3 0x4F 0x50 0x1 0x6D 0xCD 0x0 0x7F 0xB6 0xA8 0xAB 0xAB 0x55 0xA1 0xA4 0x7F 0xB3 0xB1 0xA0 0xA3 0xA4 0xA3 0xE8 0x57 0x0 0x98 0xAE 0xB4 0x7F 0xAD 0xA4 0xA4 0xA3 0x7F 0xF9 0x7F 0x54 0x8C 0x8E 0x8D 0x4F 0xB3 0xAE 0x7F 0xA5 0xA8 0xA6 0xA7 0xB3 0xE7 0x58 0x0 0x92 0xAE 0xB1 0xB1 0xB8 0xF4 0x7F 0x8C 0x84 0x96 0x7F 0xA2 0xA0 0xAD 0xBE 0x4F 0xA0 0xB3 0xB3 0xA4 0xAD 0xA3 0xE7 0x58 0x0 0x98 0xAE 0xB4 0xB1 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xAC 0xB4 0xB2 0xB3 0x4F 0xA0 0xAB 0xAB 0x7F 0xA1 0xA4 0x7F 0xA3 0xA8 0xA5 0xA5 0xA4 0xB1 0xA4 0xAD 0xB3 0xE7 0x58 0x0 0x8D 0xAE 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xA2 0xA0 0xAD 0x4F 0xA4 0xB7 0xA2 0xA4 0xA4 0xA3 0x7F 0x8B 0xFB 0xFB 0xE7 0x58 0x0 0x80 0xAB 0xAB 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xAC 0xB4 0xB2 0xB3 0x4F 0xA1 0xA4 0x7F 0xA0 0xB3 0x7F 0xAB 0xA4 0xA0 0xB2 0xB3 0x7F 0x8B 0xFB 0xF6 0xE7 0x58 0x0 0x98 0xAE 0xB4 0xB1 0x7F 0xB3 0xAE 0xB3 0xA0 0xAB 0x7F 0xAB 0xA4 0xB5 0xA4 0xAB 0xB2 0x4F 0xA4 0xB7 0xA2 0xA4 0xA4 0xA3 0x7F 0xF7 0xFB 0xFB 0xE7 0x58 0x0 0x8D 0xAE 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xA2 0xA0 0xAD 0x4F 0xA4 0xB7 0xA2 0xA4 0xA4 0xA3 0x7F 0x8B 0xF9 0xF6 0xE7 0x58 0x0 0x80 0xAB 0xAB 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xAC 0xB4 0xB2 0xB3 0x4F 0xA1 0xA4 0x7F 0xA0 0xB3 0x7F 0xAB 0xA4 0xA0 0xB2 0xB3 0x7F 0x8B 0xF8 0xFB 0xE7 0x58 0x0 0x98 0xAE 0xB4 0xB1 0x7F 0xB3 0xAE 0xB3 0xA0 0xAB 0x7F 0xAB 0xA4 0xB5 0xA4 0xAB 0xB2 0x4F 0xA4 0xB7 0xA2 0xA4 0xA4 0xA3 0x7F 0xFE 0xF6 0xE7 0x58 0x0 0x8D 0xAE 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xA2 0xA0 0xAD 0x4F 0xA4 0xB7 0xA2 0xA4 0xA4 0xA3 0x7F 0x8B 0xF8 0xF6 0xE7 0x58 0x0 0x80 0xAB 0xAB 0x7F 0x54 0x8C 0x8E 0x8D 0x7F 0xAC 0xB4 0xB2 0xB3 0x4F 0xA1 0xA4 0x7F 0xA0 0xB3 0x7F 0xAB 0xA4 0xA0 0xB2 0xB3 0x7F 0x8B 0xF7 0xFB 0xE7 0x58 0x0 0x98 0xAE 0xB4 0xB1 0x7F 0xB3 0xAE 0xB3 0xA0 0xAB 0x7F 0xAB 0xA4 0xB5 0xA4 0xAB 0xB2 0x4F 0xA4 0xB7 0xA2 0xA4 0xA4 0xA3 0x7F 0xFB 0xF6 0xE7 0x58 0x1 0x6D 0xCD 0x0 0x7F 0xA8 0xB2 0x7F 0xAE 0xB5 0xA4 0xB1 0x4F 0xFC 0x71 0xFE 0x73 0x7F 0xB3 0xA0 0xAB 0xAB 0xE7 0x58 0x1 0x6D 0xCD 0x0)
                 ;; (list
                 ;;  "Sleeping on the"
                 ;;  0x4F
                 ;;  "job again are we?"
                 ;;  0x51
                 ;;  0x00
                 ;;  "test"
                 ;;  0x59
                 ;;  "EOM")

 
                  )
     (int-array)
     (write-rom!))
    (root)
     
    )
  
  )

(comment
"Hello there![0x4F]Welcome to the[0x55]world of [POKE]MON![0x51]My
name is OAK![0x4F]People call me[0x55]the [POKE]MON PROF![0x58]This
world is[0x4F]inhabited by[0x55]creatures
called[0x55][POKE]MON![0x50][0x50][0x51]For some people,
[0x4F][POKE]MON are[0x55]pets. Others use[0x55]them for
fights.[0x51]Myself...[0x51]I study [POKE]MON[0x4F]as a
profession.[0x58]First,  what is[0x4F]your name?[0x58]This is my
grand-[0x4F]son. He's been[0x55]your rival since[0x55]you were a
baby.[0x51]...Erm,  what is[0x4F]his name again?[0x58][RED]![0x51]Your
very own[0x4F][POKE]MON legend is[0x55]about to unfold![0x51]A world
of dreams[0x4F]and adventures[0x55]with [POKE]MON[0x55]awaits! Let's
go![0x57]")













